#include "config.h"

#include "auto-boot-check.h"
#include "abc-utils.h"

#include <ifaddrs.h>
#include <arpa/inet.h>
#include <net/if.h>

#define autofreeifaddrs __attribute__((cleanup(cleanup_freeifaddrs)))

static inline void cleanup_freeifaddrs(void *p)
{
	void **pp = (void **)p;

	if (*pp != NULL)
		freeifaddrs(*pp);
}

static uint32_t prefix_len_to_netmask(int prefix_len)
{
	return htonl(0xffffffff << (32 - prefix_len));
}

int check_ip(const char **args, arg_unused int n_args)
{
	const char *ip = args[0];
	const char *prefix_len_s = args[1];

	struct in_addr ip_addr;
	if (!inet_pton(AF_INET, ip, &ip_addr)) {
		error("Invalid ip arg\n");
		return 1;
	}

	int prefix_len = atoi(prefix_len_s);
	if (prefix_len < 0 || prefix_len > 32) {
		error("Invalid prefix length\n");
		return 1;
	}

	uint32_t netmask_addr = prefix_len_to_netmask(prefix_len);

	autofreeifaddrs struct ifaddrs *ifaddr = NULL;
	if (getifaddrs(&ifaddr) == -1) {
		error("getifaddrs failed: %s", strerror(errno));
		return 1;
	}

	for (struct ifaddrs *ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
		if (ifa->ifa_addr == NULL)
			continue;

		if (ifa->ifa_addr->sa_family == AF_INET) {
			if (ifa->ifa_flags & IFF_LOOPBACK)
				continue;

			if (!(ifa->ifa_flags & IFF_RUNNING))
				continue;

			struct sockaddr_in *if_ip =
				(struct sockaddr_in *)ifa->ifa_addr;
			if (if_ip->sin_addr.s_addr != ip_addr.s_addr)
				continue;

			/* Found the ip, now verify the prefix len */

			struct sockaddr_in *if_nm =
				(struct sockaddr_in *)ifa->ifa_netmask;
			if (if_nm->sin_addr.s_addr != netmask_addr) {
				error("Wrong netmask for interface with ip %s\n", ip);
				return 1;
			}

			return 0;
		}
	}

	error("No interface with ip %s found\n", ip);
	return 1;
}

int check_default_gw(const char **args, arg_unused int n_args)
{
	const char *expected_gw = args[0];

	struct in_addr expected_gw_addr;
	if (!inet_pton(AF_INET, expected_gw, &expected_gw_addr)) {
		error("Invalid gw arg\n");
		return 1;
	}

	autofree char *route_conf = read_file("/proc/net/route");
	if (route_conf == NULL)
		fatal("Failed to read /proc/net/route\n");

	size_t n_rows;
	autofree char **rows = str_split(route_conf, '\n', &n_rows);

	// Start at 1 to skip header
	for (size_t i = 1; i < n_rows; i++) {
		char *row = rows[i];
		if (*row == 0)
			continue;

		size_t n_fields;
		autofree char **fields = str_split(row, '\t', &n_fields);

		if (n_fields < 3) {
			error("Invalid route entry\n");
			continue;
		}

		char *dest_field = fields[1];
		char *gw_field = fields[2];

		if (strcmp(dest_field, "00000000") == 0) {
			long real_gw_addr = strtol(gw_field, NULL, 16);

			if (real_gw_addr == expected_gw_addr.s_addr)
				return 0;
			else {
				error("Wrong default gateway\n");
				return 1;
			}
		}
	}

	error("Unable to find default gateway\n");
	return 1;
}

int check_nameserver(const char **args, arg_unused int n_args)
{
	const char *expected_nameserver = args[0];

	autofree char *resolv_conf = read_file("/etc/resolv.conf");
	if (resolv_conf == NULL)
		fatal("Failed to read resolv.conf\n");

	size_t n_rows;
	autofree char **rows = str_split(resolv_conf, '\n', &n_rows);

	for (size_t i = 0; i < n_rows; i++) {
		char *row = rows[i];
		if (*row == '#')
			continue;

		while (*row == ' ')
			row++;

		if (str_has_prefix(row, "nameserver ")) {
			char *ns = row + strlen("nameserver");
			while (*ns == ' ')
				ns++;

			if (strcmp(ns, expected_nameserver) == 0)
				return 0;

			error("Wrong nameserver %s configured, expected: %s\n",
			      ns, expected_nameserver);
			return 1;
		}
	}

	error("Unable to find configured nameserver\n");
	return 1;
}
