#include "config.h"

#include "abc-utils.h"

void verbose(const char *format, ...)
{
	if (opt_verbose) {
		va_list args;
		va_start(args, format);
		vprintf(format, args);
		va_end(args);
	}
}

void error(const char *format, ...)
{
	va_list args;

	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
}

void fatal(const char *format, ...)
{
	va_list args;

	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);

	exit(1);
}

void oom()
{
	fatal("Out of memory");
}

void *xmalloc(size_t size)
{
	void *res = malloc(size);
	if (res == NULL)
		oom();
	return res;
}

void *xcalloc(size_t nmemb, size_t size)
{
	void *res = calloc(nmemb, size);
	if (res == NULL)
		oom();
	return res;
}

void *xrealloc(void *ptr, size_t size)
{
	void *res = realloc(ptr, size);
	if (res == NULL)
		oom();
	return res;
}

bool str_has_prefix(const char *str, const char *prefix)
{
	return strncmp(str, prefix, strlen(prefix)) == 0;
}

bool str_has_suffix(const char *str, const char *suffix)
{
	size_t str_len;
	size_t suffix_len;

	str_len = strlen(str);
	suffix_len = strlen(suffix);

	if (str_len < suffix_len)
		return false;

	return strcmp(str + str_len - suffix_len, suffix) == 0;
}

char *join_path(const char *a, const char *b)
{
	size_t a_len = strlen(a);
	size_t b_len = strlen(b);

	bool add_slash = a_len > 0 && a[a_len - 1] != '/';

	size_t len = a_len + b_len + 1;
	if (add_slash)
		len++;

	char *res = xmalloc(len);

	size_t res_size = a_len;
	memcpy(res, a, a_len);

	if (add_slash) {
		memcpy(res + res_size, "/", 1);
		res_size += 1;
	}
	memcpy(res + res_size, b, b_len + 1);

	return res;
}

/* This is used by the tests */
static char *rewrite_root_prefix(char *path)
{
	static const char *root_prefix = NULL;
	static bool checked_root_prefix = false;

	if (!checked_root_prefix) {
		root_prefix = getenv("AUTO_BOOT_CHECK_ROOT_PREFIX");
		checked_root_prefix = true;
	}

	if (root_prefix) {
		char *rewritten = join_path(root_prefix, path);
		free(path);
		path = rewritten;
	}

	return path;
}

char *read_file(const char *filename)
{
	autoclose int fd = -1;
	size_t n_read;
	size_t buf_size;

	fd = open(filename, O_RDONLY);
	if (fd == -1)
		return NULL;

	buf_size = 4096;
	autofree char *buf = xmalloc(buf_size);

	n_read = 0;
	while (1) {
		if (n_read == buf_size) {
			buf_size *= 2;
			buf = xrealloc(buf, buf_size);
		}

		ssize_t r = read(fd, buf + n_read, buf_size - n_read);
		if (r < 0)
			return NULL;

		n_read += r;
		if (r == 0) /* early eof */
			break;
	}

	// Always null-terminate
	buf[n_read] = 0;

	return steal_ptr(&buf);
}

/* Note: Modifies row */
char **str_split(char *row, char split_at, size_t *size)
{
	size_t n;
	char **res;

	n = 1;
	for (int i = 0; row[i] != 0; i++) {
		if (row[i] == split_at)
			n++;
	}

	res = xcalloc(n + 1, sizeof(char *));

	n = 0;
	res[n++] = row;
	while (*row != 0) {
		while (*row != 0 && *row != split_at)
			row++;
		if (*row == split_at) {
			*row = 0;
			row++;
			res[n++] = row;
		}
	}

	if (size)
		*size = n;

	res[n] = NULL; /* NULL terminate */

	return res;
}

/* Given a json object such as found in
 * /var/lib/containers/storage/overlay-containers/containers.json,
 * find the container json object in it with a given name */
static struct json_object *containers_info_find_by_name(struct json_object *containers,
							const char *container_name)
{
	if (containers == NULL || !json_object_is_type(containers, json_type_array))
		return NULL;

	for (size_t i = 0; i < json_object_array_length(containers); i++) {
		struct json_object *container =
			json_object_array_get_idx(containers, i);
		struct json_object *names;

		if (!json_object_object_get_ex(container, "names", &names) ||
		    !json_object_is_type(names, json_type_array))
			continue;

		for (size_t j = 0; j < json_object_array_length(names); j++) {
			struct json_object *name =
				json_object_array_get_idx(names, j);

			if (json_object_is_type(name, json_type_string) &&
			    strcmp(json_object_get_string(name), container_name) == 0) {
				return container;
			}
		}
	}

	return NULL;
}

/* Look up the information from the containers.json file
 * in the podman local storage for a named container, or
 * return NULL if no such container found. */
struct json_object *find_container_info_by_name(const char *container_name)
{
	/* We hardcode the standard locations and the overlay backend,
	 * but we look for both regular and volatile containers. */
	const char *dirs[] = {
		"/var/lib/containers/storage/",
		"/run/containers/storage/",
	};
	const char *files[] = {
		"overlay-containers/containers.json",
		"overlay-containers/volatile-containers.json",
	};

	for (size_t i = 0; i < N_ELEMENTS(dirs); i++) {
		for (size_t j = 0; j < N_ELEMENTS(files); j++) {
			autofree char *path =
				rewrite_root_prefix(join_path(dirs[i], files[j]));
			autofreejson json_object *containers =
				json_object_from_file(path);
			if (containers) {
				struct json_object *container_info =
					containers_info_find_by_name(
						containers, container_name);
				if (container_info)
					return json_object_get(container_info);
			}
		}
	}

	return NULL;
}

/* Look up the container id from a container info object, such as
 * those in containers.json */
const char *container_info_get_id(struct json_object *container_info)
{
	struct json_object *id;
	if (container_info != NULL &&
	    json_object_object_get_ex(container_info, "id", &id) &&
	    json_object_is_type(id, json_type_string)) {
		return json_object_get_string(id);
	}

	return NULL;
}

/* Given a container id, look at running containers and lookup the
 * OCI config file generated for it. */
struct json_object *load_container_config(const char *container_id)
{
	/* We hardcode the standard locations. */
	const char *dirs[] = {
		"/var/lib/containers/storage/overlay-containers",
		"/run/containers/storage/overlay-containers",
	};

	for (size_t i = 0; i < N_ELEMENTS(dirs); i++) {
		autofree char *id_dir =
			rewrite_root_prefix(join_path(dirs[i], container_id));
		autofree char *config_path =
			join_path(id_dir, "userdata/config.json");
		autofreejson json_object *config =
			json_object_from_file(config_path);
		if (config) {
			return json_object_get(config);
		}
	}

	return NULL;
}

/* If there are an instances of `from` in the string, return a new
 * string that replace all instances with `to`. If there are no such
 * instances return NULL, allowing reuse the original string */
static char *json_maybe_rewrite_str(const char *str, const char *from, const char *to)
{
	int n_matches = 0;

	char *found = strstr(str, from);
	while (found != NULL) {
		n_matches++;
		found = strstr(found + strlen(from), from);
	}
	if (n_matches == 0)
		return NULL;

	char *res = xmalloc(strlen(str) + n_matches * strlen(to) -
			    n_matches * strlen(from) + 1);

	char *out = res;
	found = strstr(str, from);
	while (found != NULL) {
		memcpy(out, str, found - str);
		out += found - str;
		memcpy(out, to, strlen(to));
		out += strlen(to);
		str = found + strlen(from);
		found = strstr(str, from);
	}
	memcpy(out, str, strlen(str));
	out += strlen(str);
	*out = 0;

	return res;
}

/* Recursively rewrite all instances of `from` to `to` in all strings in the json data */
static void json_rewrite_strings(struct json_object *value, const char *from,
				 const char *to)
{
	enum json_type type = json_object_get_type(value);
	switch (type) {
	case json_type_string:
		const char *s = json_object_get_string(value);
		char *res = json_maybe_rewrite_str(s, from, to);
		if (res) {
			json_object_set_string(value, res);
			free(res);
		}

		break;
	case json_type_object:
		json_object_object_foreach(value, key, val)
		{
			struct json_object *child_value;
			json_object_object_get_ex(value, key, &child_value);
			json_rewrite_strings(child_value, from, to);
		}

		break;
	case json_type_array:
		for (size_t i = 0; i < json_object_array_length(value); i++) {
			struct json_object *child_value =
				json_object_array_get_idx(value, i);
			json_rewrite_strings(child_value, from, to);
		}
		break;
	case json_type_null:
	case json_type_boolean:
	case json_type_double:
	case json_type_int:
		break;
	}
}

static int container_config_sort_by_key(const void *_a, const void *_b,
					const char *key)
{
	struct json_object *a = *(struct json_object **)_a;
	struct json_object *b = *(struct json_object **)_b;

	struct json_object *a_dest, *b_dest;

	if (json_object_is_type(a, json_type_object) &&
	    json_object_object_get_ex(a, key, &a_dest) &&
	    json_object_is_type(a_dest, json_type_string) &&
	    json_object_is_type(b, json_type_object) &&
	    json_object_object_get_ex(b, key, &b_dest) &&
	    json_object_is_type(b_dest, json_type_string)) {
		return strcmp(json_object_get_string(a_dest),
			      json_object_get_string(b_dest));
	}

	return 0;
}

static int container_config_sort_mounts(const void *a, const void *b)
{
	return container_config_sort_by_key(a, b, "destination");
}

static int container_config_sort_namespaces(const void *a, const void *b)
{
	return container_config_sort_by_key(a, b, "type");
}

static struct json_object *container_config_get_namespace(struct json_object *config,
							  const char *find_ns)
{
	struct json_object *linux_dict;

	if (!json_object_object_get_ex(config, "linux", &linux_dict) ||
	    !json_object_is_type(linux_dict, json_type_object))
		return NULL;

	struct json_object *namespaces;
	if (!json_object_object_get_ex(linux_dict, "namespaces", &namespaces) ||
	    !json_object_is_type(namespaces, json_type_array)) {
		return NULL;
	}

	for (size_t i = 0; i < json_object_array_length(namespaces); i++) {
		struct json_object *ns = json_object_array_get_idx(namespaces, i);

		struct json_object *type;
		if (json_object_is_type(ns, json_type_object) &&
		    json_object_object_get_ex(ns, "type", &type) &&
		    json_object_is_type(type, json_type_string) &&
		    strcmp(json_object_get_string(type), find_ns) == 0) {
			return ns;
		}
	}

	return NULL;
}

/* Does various changes to an OCI container config with the intent of making it canonical.
 * These changes are chosen to undo some random effects of the podman generation of such
 * files */
void canonicalize_container_config(struct json_object *config, const char *container_id)
{
	char partial_id[33];

	/* Lots of strings container the container id of various lengths, replace theme by $CONTAINER */
	json_rewrite_strings(config, container_id, "$CONTAINER");

	strncpy(partial_id, container_id, 32);
	partial_id[32] = 0;
	json_rewrite_strings(config, partial_id, "$CONTAINER");

	partial_id[12] = 0;
	json_rewrite_strings(config, partial_id, "$CONTAINER");

	/* The some arrays seems to be of random order, so sort it */
	struct json_object *mounts;
	if (json_object_object_get_ex(config, "mounts", &mounts)) {
		json_object_array_sort(mounts, container_config_sort_mounts);
	}

	struct json_object *namespaces;
	if (json_object_object_get_ex(config, "namespaces", &namespaces)) {
		json_object_array_sort(mounts, container_config_sort_namespaces);
	}

	/* The private network namespace contains a per-instance uuid */
	struct json_object *network_ns =
		container_config_get_namespace(config, "network");
	if (network_ns) {
		struct json_object *path;
		if (json_object_object_get_ex(network_ns, "path", &path) &&
		    json_object_is_type(path, json_type_string) &&
		    str_has_prefix(json_object_get_string(path), "/run/netns/")) {
			json_object_set_string(path, "/run/netns/$NETNS");
		}
	}
}

/*
 * SHA-256 Checksum
 */

/* adapted from the SHA256 implementation in glib, which is originally:
 *
 * Copyright (C) 2006 Dave Benson
 * Released under the terms of the GNU Lesser General Public License
 */

void sha256_sum_init(Sha256sum *sha256)
{
	sha256->buf[0] = 0x6a09e667;
	sha256->buf[1] = 0xbb67ae85;
	sha256->buf[2] = 0x3c6ef372;
	sha256->buf[3] = 0xa54ff53a;
	sha256->buf[4] = 0x510e527f;
	sha256->buf[5] = 0x9b05688c;
	sha256->buf[6] = 0x1f83d9ab;
	sha256->buf[7] = 0x5be0cd19;

	sha256->bits[0] = sha256->bits[1] = 0;
}

#define GET_UINT32(n, b, i)                                                     \
	do {                                                                    \
		(n) = ((uint32_t)(b)[(i)] << 24) |                              \
		      ((uint32_t)(b)[(i) + 1] << 16) |                          \
		      ((uint32_t)(b)[(i) + 2] << 8) | ((uint32_t)(b)[(i) + 3]); \
	} while (0)

#define PUT_UINT32(n, b, i)                                                    \
	do {                                                                   \
		(b)[(i)] = (uint8_t)((n) >> 24);                               \
		(b)[(i) + 1] = (uint8_t)((n) >> 16);                           \
		(b)[(i) + 2] = (uint8_t)((n) >> 8);                            \
		(b)[(i) + 3] = (uint8_t)((n));                                 \
	} while (0)

static void sha256_transform(uint32_t buf[8], uint8_t const data[64])
{
	uint32_t temp1, temp2, W[64];
	uint32_t A, B, C, D, E, F, G, H;

	GET_UINT32(W[0], data, 0);
	GET_UINT32(W[1], data, 4);
	GET_UINT32(W[2], data, 8);
	GET_UINT32(W[3], data, 12);
	GET_UINT32(W[4], data, 16);
	GET_UINT32(W[5], data, 20);
	GET_UINT32(W[6], data, 24);
	GET_UINT32(W[7], data, 28);
	GET_UINT32(W[8], data, 32);
	GET_UINT32(W[9], data, 36);
	GET_UINT32(W[10], data, 40);
	GET_UINT32(W[11], data, 44);
	GET_UINT32(W[12], data, 48);
	GET_UINT32(W[13], data, 52);
	GET_UINT32(W[14], data, 56);
	GET_UINT32(W[15], data, 60);

#define SHR(x, n) ((x & 0xFFFFFFFF) >> n)
#define ROTR(x, n) (SHR(x, n) | (x << (32 - n)))

#define S0(x) (ROTR(x, 7) ^ ROTR(x, 18) ^ SHR(x, 3))
#define S1(x) (ROTR(x, 17) ^ ROTR(x, 19) ^ SHR(x, 10))
#define S2(x) (ROTR(x, 2) ^ ROTR(x, 13) ^ ROTR(x, 22))
#define S3(x) (ROTR(x, 6) ^ ROTR(x, 11) ^ ROTR(x, 25))

#define F0(x, y, z) ((x & y) | (z & (x | y)))
#define F1(x, y, z) (z ^ (x & (y ^ z)))

#define R(t) (W[t] = S1(W[t - 2]) + W[t - 7] + S0(W[t - 15]) + W[t - 16])

#define P(a, b, c, d, e, f, g, h, x, K)                                        \
	do {                                                                   \
		temp1 = h + S3(e) + F1(e, f, g) + K + x;                       \
		temp2 = S2(a) + F0(a, b, c);                                   \
		d += temp1;                                                    \
		h = temp1 + temp2;                                             \
	} while (0)

	A = buf[0];
	B = buf[1];
	C = buf[2];
	D = buf[3];
	E = buf[4];
	F = buf[5];
	G = buf[6];
	H = buf[7];

	P(A, B, C, D, E, F, G, H, W[0], 0x428A2F98);
	P(H, A, B, C, D, E, F, G, W[1], 0x71374491);
	P(G, H, A, B, C, D, E, F, W[2], 0xB5C0FBCF);
	P(F, G, H, A, B, C, D, E, W[3], 0xE9B5DBA5);
	P(E, F, G, H, A, B, C, D, W[4], 0x3956C25B);
	P(D, E, F, G, H, A, B, C, W[5], 0x59F111F1);
	P(C, D, E, F, G, H, A, B, W[6], 0x923F82A4);
	P(B, C, D, E, F, G, H, A, W[7], 0xAB1C5ED5);
	P(A, B, C, D, E, F, G, H, W[8], 0xD807AA98);
	P(H, A, B, C, D, E, F, G, W[9], 0x12835B01);
	P(G, H, A, B, C, D, E, F, W[10], 0x243185BE);
	P(F, G, H, A, B, C, D, E, W[11], 0x550C7DC3);
	P(E, F, G, H, A, B, C, D, W[12], 0x72BE5D74);
	P(D, E, F, G, H, A, B, C, W[13], 0x80DEB1FE);
	P(C, D, E, F, G, H, A, B, W[14], 0x9BDC06A7);
	P(B, C, D, E, F, G, H, A, W[15], 0xC19BF174);
	P(A, B, C, D, E, F, G, H, R(16), 0xE49B69C1);
	P(H, A, B, C, D, E, F, G, R(17), 0xEFBE4786);
	P(G, H, A, B, C, D, E, F, R(18), 0x0FC19DC6);
	P(F, G, H, A, B, C, D, E, R(19), 0x240CA1CC);
	P(E, F, G, H, A, B, C, D, R(20), 0x2DE92C6F);
	P(D, E, F, G, H, A, B, C, R(21), 0x4A7484AA);
	P(C, D, E, F, G, H, A, B, R(22), 0x5CB0A9DC);
	P(B, C, D, E, F, G, H, A, R(23), 0x76F988DA);
	P(A, B, C, D, E, F, G, H, R(24), 0x983E5152);
	P(H, A, B, C, D, E, F, G, R(25), 0xA831C66D);
	P(G, H, A, B, C, D, E, F, R(26), 0xB00327C8);
	P(F, G, H, A, B, C, D, E, R(27), 0xBF597FC7);
	P(E, F, G, H, A, B, C, D, R(28), 0xC6E00BF3);
	P(D, E, F, G, H, A, B, C, R(29), 0xD5A79147);
	P(C, D, E, F, G, H, A, B, R(30), 0x06CA6351);
	P(B, C, D, E, F, G, H, A, R(31), 0x14292967);
	P(A, B, C, D, E, F, G, H, R(32), 0x27B70A85);
	P(H, A, B, C, D, E, F, G, R(33), 0x2E1B2138);
	P(G, H, A, B, C, D, E, F, R(34), 0x4D2C6DFC);
	P(F, G, H, A, B, C, D, E, R(35), 0x53380D13);
	P(E, F, G, H, A, B, C, D, R(36), 0x650A7354);
	P(D, E, F, G, H, A, B, C, R(37), 0x766A0ABB);
	P(C, D, E, F, G, H, A, B, R(38), 0x81C2C92E);
	P(B, C, D, E, F, G, H, A, R(39), 0x92722C85);
	P(A, B, C, D, E, F, G, H, R(40), 0xA2BFE8A1);
	P(H, A, B, C, D, E, F, G, R(41), 0xA81A664B);
	P(G, H, A, B, C, D, E, F, R(42), 0xC24B8B70);
	P(F, G, H, A, B, C, D, E, R(43), 0xC76C51A3);
	P(E, F, G, H, A, B, C, D, R(44), 0xD192E819);
	P(D, E, F, G, H, A, B, C, R(45), 0xD6990624);
	P(C, D, E, F, G, H, A, B, R(46), 0xF40E3585);
	P(B, C, D, E, F, G, H, A, R(47), 0x106AA070);
	P(A, B, C, D, E, F, G, H, R(48), 0x19A4C116);
	P(H, A, B, C, D, E, F, G, R(49), 0x1E376C08);
	P(G, H, A, B, C, D, E, F, R(50), 0x2748774C);
	P(F, G, H, A, B, C, D, E, R(51), 0x34B0BCB5);
	P(E, F, G, H, A, B, C, D, R(52), 0x391C0CB3);
	P(D, E, F, G, H, A, B, C, R(53), 0x4ED8AA4A);
	P(C, D, E, F, G, H, A, B, R(54), 0x5B9CCA4F);
	P(B, C, D, E, F, G, H, A, R(55), 0x682E6FF3);
	P(A, B, C, D, E, F, G, H, R(56), 0x748F82EE);
	P(H, A, B, C, D, E, F, G, R(57), 0x78A5636F);
	P(G, H, A, B, C, D, E, F, R(58), 0x84C87814);
	P(F, G, H, A, B, C, D, E, R(59), 0x8CC70208);
	P(E, F, G, H, A, B, C, D, R(60), 0x90BEFFFA);
	P(D, E, F, G, H, A, B, C, R(61), 0xA4506CEB);
	P(C, D, E, F, G, H, A, B, R(62), 0xBEF9A3F7);
	P(B, C, D, E, F, G, H, A, R(63), 0xC67178F2);

#undef SHR
#undef ROTR
#undef S0
#undef S1
#undef S2
#undef S3
#undef F0
#undef F1
#undef R
#undef P

	buf[0] += A;
	buf[1] += B;
	buf[2] += C;
	buf[3] += D;
	buf[4] += E;
	buf[5] += F;
	buf[6] += G;
	buf[7] += H;
}

void sha256_sum_update(Sha256sum *sha256, const uint8_t *buffer, size_t length)
{
	uint32_t left, fill;
	const uint8_t *input = buffer;

	if (length == 0)
		return;

	left = sha256->bits[0] & 0x3F;
	fill = 64 - left;

	sha256->bits[0] += length;
	sha256->bits[0] &= 0xFFFFFFFF;

	if (sha256->bits[0] < length)
		sha256->bits[1]++;

	if (left > 0 && length >= fill) {
		memcpy((sha256->data + left), input, fill);

		sha256_transform(sha256->buf, sha256->data);
		length -= fill;
		input += fill;

		left = 0;
	}

	while (length >= SHA256_DATASIZE) {
		sha256_transform(sha256->buf, input);

		length -= 64;
		input += 64;
	}

	if (length)
		memcpy(sha256->data + left, input, length);
}

static uint8_t sha256_padding[64] = { 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				      0,    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				      0,    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				      0,    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				      0,    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

void sha256_sum_close(Sha256sum *sha256)
{
	uint32_t last, padn;
	uint32_t high, low;
	uint8_t msglen[8];

	high = (sha256->bits[0] >> 29) | (sha256->bits[1] << 3);
	low = (sha256->bits[0] << 3);

	PUT_UINT32(high, msglen, 0);
	PUT_UINT32(low, msglen, 4);

	last = sha256->bits[0] & 0x3F;
	padn = (last < 56) ? (56 - last) : (120 - last);

	sha256_sum_update(sha256, sha256_padding, padn);
	sha256_sum_update(sha256, msglen, 8);

	PUT_UINT32(sha256->buf[0], sha256->digest, 0);
	PUT_UINT32(sha256->buf[1], sha256->digest, 4);
	PUT_UINT32(sha256->buf[2], sha256->digest, 8);
	PUT_UINT32(sha256->buf[3], sha256->digest, 12);
	PUT_UINT32(sha256->buf[4], sha256->digest, 16);
	PUT_UINT32(sha256->buf[5], sha256->digest, 20);
	PUT_UINT32(sha256->buf[6], sha256->digest, 24);
	PUT_UINT32(sha256->buf[7], sha256->digest, 28);
}

#undef PUT_UINT32
#undef GET_UINT32

void sha256_digest_to_hex(const uint8_t *csum, char *buf)
{
	static const char hexchars[] = "0123456789abcdef";
	uint32_t i, j;

	for (i = 0, j = 0; i < SHA256_DIGEST_LEN; i++, j += 2) {
		uint8_t byte = csum[i];
		buf[j] = hexchars[byte >> 4];
		buf[j + 1] = hexchars[byte & 0xF];
	}
	buf[j] = '\0';
}

void sha256_hex(const uint8_t *buffer, size_t length, char *out_buf)
{
	Sha256sum sum;
	sha256_sum_init(&sum);
	sha256_sum_update(&sum, buffer, length);

	sha256_sum_close(&sum);

	sha256_digest_to_hex(sum.digest, out_buf);
}
