#include "config.h"

#include <getopt.h>

#include "auto-boot-check.h"
#include "abc-utils.h"

/* Global state */
bool opt_verbose = 0;

int container_checksum(int argc, char *argv[])
{
	if (argc != 1) {
		error("No container name specified\n");
		return 1;
	}

	const char *container_name = argv[0];

	autofreejson struct json_object *container_info =
		find_container_info_by_name(container_name);

	if (container_info == NULL) {
		error("Container named '%s' not found\n", container_name);
		return 1;
	}

	const char *container_id = container_info_get_id(container_info);
	if (container_id == NULL) {
		error("Unable to get container id\n");
		return 1;
	}

	autofreejson struct json_object *config =
		load_container_config(container_id);
	if (config == NULL) {
		error("Unable to get container config\n");
		return 1;
	}

	canonicalize_container_config(config, container_id);

	const char *js = json_object_to_json_string(config);
	char digest_hex[SHA256_DIGEST_HEX_LEN + 1];
	sha256_hex((uint8_t *)js, strlen(js), digest_hex);

	printf("%s\n", digest_hex);

	return 0;
}

#define OPT_VERSION 102
#define OPT_HELP 103
#define OPT_VERBOSE 104

static void usage(void)
{
	fprintf(stderr,
		"Usage: auto-boot-check-tool [OPTION]  COMMAND [...]\n"
		"Options:\n"
		"  --version             Show version\n"
		"  --help                Show help\n"
		"Commands:\n"
		"  container-checksum NAME    Compute checksum of named container\n");
}

int main(int argc, char *argv[])
{
	const struct option longopts[] = {
		{ .name = "help", .has_arg = no_argument, .flag = NULL, .val = OPT_HELP },
		{ .name = "verbose", .has_arg = no_argument, .flag = NULL, .val = OPT_VERBOSE },
		{ .name = "version", .has_arg = no_argument, .flag = NULL, .val = OPT_VERSION },
		{},
	};
	int opt;

	while ((opt = getopt_long(argc, argv, "", longopts, NULL)) != -1) {
		switch (opt) {
		case OPT_HELP:
			usage();
			exit(0);
			break;
		case OPT_VERSION:
			printf("auto-boot-check %s\n", PACKAGE_VERSION);
			exit(0);
			break;
		case OPT_VERBOSE:
			opt_verbose = true;
			break;
		case ':':
			fprintf(stderr, "option needs a value\n");
			exit(EXIT_FAILURE);
		default:
			usage();
			exit(1);
		}
	}

	argv += optind;
	argc -= optind;

	if (argc < 1) {
		error("No command specified\n");
		usage();
		return 1;
	}

	const char *command = argv[0];

	if (strcmp(command, "container-checksum") == 0) {
		return container_checksum(argc - 1, argv + 1);
	}

	error("Unsupported command '%s'\n", command);
	usage();

	return 1;
}
