#include "config.h"

#include "auto-boot-check.h"
#include "abc-utils.h"

/* Global state */
bool opt_verbose = 0;

typedef struct {
	char *name;
	size_t min_args;
	int (*fn)(const char **args, int n_args);
} CmdHandler;

CmdHandler commands[] = {
	{ "ip", 2, check_ip },
	{ "default_gw", 1, check_default_gw },
	{ "nameserver", 1, check_nameserver },
	{ "container_checksum", 2, check_container_checksum },
};

static CmdHandler *find_command_handler(char *name)
{
	for (size_t i = 0; i < N_ELEMENTS(commands); i++) {
		CmdHandler *handler = &commands[i];
		if (strcmp(name, handler->name) == 0)
			return handler;
	}

	return NULL;
}

static int run_file_checks(const char *filename)
{
	autofree char *check_data = NULL;

	verbose("Reading checks from %s\n", filename);

	check_data = read_file(filename);
	if (check_data == NULL) {
		error("Couldn't open file %s\n", filename);
		return 1;
	}

	size_t n_rows;
	autofree char **rows = str_split(check_data, '\n', &n_rows);

	int res = 0;

	for (size_t i = 0; i < n_rows; i++) {
		char *row = rows[i];
		autofree char **args = NULL;
		size_t n_args;

		if (strlen(row) == 0) // Skip trailing row
			continue;

		if (row[0] == '#') // Skip comment
			continue;

		args = str_split(row, ' ', &n_args);

		if (n_args == 0)
			continue;

		char *cmd_name = args[0];
		CmdHandler *handler = find_command_handler(cmd_name);

		if (handler == NULL) {
			error("Unsupported command '%s'\n", cmd_name);
			res = 1;
			continue;
		}

		if (n_args - 1 < handler->min_args) {
			error("Not enough args to command '%s'\n", cmd_name);
			res = 1;
			continue;
		}

		if (opt_verbose) {
			verbose("Checking: %s", cmd_name);
			for (size_t i = 1; i < n_args; i++)
				verbose(" %s", args[i]);
			verbose("\n");
		}

		if (handler->fn(((const char **)args) + 1, n_args - 1)) {
			verbose("Check failed\n");
			res = 1;
			continue;
		}
	}

	return res;
}

int run_file_or_dir_checks(const char *path)
{
	int r;
	struct stat statbuf;

	r = stat(path, &statbuf);
	if (r != 0) {
		/* Non-existing file is ok */
		if (errno == ENOENT)
			return 0;

		error("Unable to load config from %s\n", path);
		return 1;
	}

	int res = 0;

	if (S_ISDIR(statbuf.st_mode)) {
		autoclosedir DIR *dir = opendir(path);
		if (!dir) {
			error("failed to open directory %s\n", path);
			return 1;
		}

		while (1) {
			errno = 0;
			struct dirent *d = readdir(dir);
			if (!d) {
				if (errno) {
					error("failed to read directory %s\n", path);
					res = 1;
				}
				break;
			}

			if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
				continue;

			if (!str_has_suffix(d->d_name, ".conf"))
				continue;

			autofree char *full_path = join_path(path, d->d_name);

			if (run_file_checks(full_path)) {
				res = 1;
			}
		}
	} else {
		res = run_file_checks(path);
	}

	return res;
}

int main(int argc, char *argv[])
{
	int start_arg;

	for (start_arg = 1; start_arg < argc; start_arg++) {
		if (strcmp(argv[start_arg], "-v") == 0)
			opt_verbose = 1;
		else
			break;
	}

	int res = 0;

	if (start_arg == argc) {
		verbose("Reading config files from %s\n", CONFDIR_ETC);
		if (run_file_or_dir_checks(CONFDIR_ETC)) {
			res = 1;
		}
		verbose("Reading config files from %s\n", CONFDIR_USR);
		if (run_file_or_dir_checks(CONFDIR_USR)) {
			res = 1;
		}
	} else {
		for (int i = start_arg; i < argc; i++) {
			if (run_file_or_dir_checks(argv[i])) {
				res = 1;
			}
		}
	}

	return res;
}
